

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <map>
#include <algorithm>
#include <bitset>
#include "plink.h"
#include "options.h"
#include "helper.h"


void Plink::mergeBinaryData()
{
  
  // Function to merge a text file with an exisiting data set
  // either SNP-major or individual-major modes
  
  if (!par::merge_list)
    {
      printLOG( "Using merge mode " + int2str( par::merge_mode ) + " : ");
      if (par::merge_mode==1) printLOG("consensus call (default)\n");
      else if (par::merge_mode==2) printLOG("overwrite if missing in original\n");
      else if (par::merge_mode==3) printLOG("overwrite unless missing in new\n");
      else if (par::merge_mode==4) printLOG("overwrite none\n"); 
      else if (par::merge_mode==5) printLOG("overwrite all\n");
      else if (par::merge_mode==6) printLOG("diff mode: all differences\n");
      else if (par::merge_mode==7) printLOG("diff mode: non-missing differences\n");
    }
  
  // We've already loaded in the first file
  // Do not overwrite any existing phenotype information
  
  checkFileExists(par::merge_bedfile);
  checkFileExists(par::merge_bimfile);
  checkFileExists(par::merge_famfile);
  
  // Make hash of original SNP names
  
  map<string,int> mlocus;
  for (int l=0;l<nl_all;l++)
    mlocus.insert(make_pair(locus[l]->name,l));
  map<string,int>::iterator ilocus;
  
  
  // A temporary hash for the names of any markers that
  // do not match in terms of strand
  map<string,int> misstrand;


  ///////////////////////////////////////
  // .bim 

  vector<Locus> ordered(0);
  map<string,bool> exists;
  vector<Locus*> locus2(0);
  
  ifstream MAP(par::merge_bimfile.c_str(), ios::in);
  MAP.clear();

  int exist_cnt=0;
  
  int c=0;
  while(!MAP.eof())
    {
      
      Locus * loc = new Locus;
      
      long int inc;

      MAP >> loc->chr   // will automatically by numeric
	    >> loc->name 
	    >> loc->pos   // will automatically be in M units
	    >> loc->bp     
	    >> loc->allele1
	    >> loc->allele2;

                  
      inc = loc->bp;
            
      // Use the frequency slot temporarily to 
      // store order information
      loc->freq = c++;
      
      // Check that cM/M specification looks correct, if 
      // we want to perform a plink-based analysis
      if (par::plink && (!par::cm_map) && (loc->pos > 50) )
	error("Looks like you need to specify --cm ??");
      
      // Convert cM to M map distances
      if (par::cm_map) loc->pos /= 100;

      // Including all loci in merge-mode
      if (loc->name!="") 
	{
	      ilocus = mlocus.find(loc->name);
	      
	      // Check whether or not this Locus already exists?
	      if (ilocus != mlocus.end())
		{
		  
		  Locus * loc2 = locus[ilocus->second];
		  
		  // Check same chromosome and positions, etc
		  if ( loc2->chr != loc->chr )
		    {
		      cerr << "Warning: different chromosome specified for " 
			   << loc->name << "\n";
		      loc->chr = loc2->chr;
		    }

		  // Check same chromosome and positions, etc
		  if ( loc2->bp != loc->bp )
		    {
		      cerr << "Warning: different physical position specified for " 
			   << loc->name << "\n";
		      loc->bp = loc2->bp;		  
		    }

		  if ( loc2->pos != loc->pos )
		    {
		      cerr << "Warning: different genetic position specified for " 
			   << loc->name << "\n";
		      loc->pos = loc2->pos;		  
		    }

		  exists[loc->name] = true;
		  exist_cnt++;
		  locus2.push_back(loc);

		}
	      else 
		{
		  // Locus does not exist -- add to locus list
		  exists[loc->name] = false;
		  locus2.push_back(loc);
		}
	      
	    

	  ordered.push_back(*loc);
	}
      
      
    }

  MAP.clear();
  MAP.close();
  
  if (!par::merge_list)
    {
      printLOG("\n" +int2str(locus2.size()) + 
	       " markers to be merged from [ " +par::merge_bimfile + " ]\n");
      printLOG("Of these, "+int2str( locus2.size()-exist_cnt ) + " are new, " +
	       int2str( exist_cnt ) +  " already exist in current data\n");
    }



  ///////////////////////////////////////////////
  // Build ordered table, so that genotypes can be inserted 
  // in correct order; then swap locus file over
  
  // Sorting a vector of pointers, so we need this special fix
  stable_sort(locus2.begin(),locus2.end(),less<Locus*>());
  
  // Sorting a normal vector
  stable_sort(ordered.begin(),ordered.end());
  
  c=0;
  for (int i=0; i<ordered.size(); i++)
    {
      // swap file order into locus position
      // and make all same chromosome
      ordered[i].bp = (int)ordered[i].freq;
      ordered[i].chr = 1;
      
	// keep track of genetic order, but only 
	// for nonmissing loci
	ordered[i].freq = c++;
    }
  
  // resort to get lookup table
  stable_sort(ordered.begin(),ordered.end());  

  // i.e. for file position k, the locus position is ordered[k]->freq

  // p2 p3 p1 p5 p4   : genetic position
  // 0  1  2  3  4    : file order

  // sort by cM
  // p1 p2 p3 p4 p5   : genetic 
  // 2  0  1  4  3    : file order
  // 0  1     2       : add genetic order: nonmissing...
  // 

  // sort by file order again
  // p2 p3 p1 p5 p4   : genetic
  // 0  1  2  3  4    : file
  // 1     0     2    : position to put in locus[l]
  

  // Add new locus2() to end of locus()
  for (int l=0; l<locus2.size(); l++)
    {
      // need to add a new MAP entry
      if (!exists.find(locus2[l]->name)->second )
	{
	  Locus * loc = new Locus;
	  loc = locus2[l];
	  locus.push_back(loc);
	}
    }
  

  
  ///////////////////////////////////////////////
  // .fam
  
  // Make new hash of Locus names
  
  mlocus.clear();
  for (int l=0;l<locus.size();l++)
    {
      mlocus.insert(make_pair(locus[l]->name,l));
    }
  if (mlocus.size() != locus.size() ) 
    {
      cerr << "Problem encountered merging files, with the following markers:\n";
       mlocus.clear();
       for (int l=0;l<locus.size();l++)
 	{
 	  if (mlocus.find(locus[l]->name) != mlocus.end())
 	    cerr << locus[l]->name << "\n";
 	  mlocus.insert(make_pair(locus[l]->name,l));
 	}
       cerr <<  "[ dump info: sizes = " << mlocus.size() << " and " << locus.size() << " ]\n";
       error("Cannot merge files. Check your MAP files.");
    } 
  
  // Make hash of existing individuals
  map<string,int> msample;
  for (int i=0;i<n; i++)
    msample.insert(make_pair(sample[i]->fid+"_"+sample[i]->iid,i));
  map<string,int>::iterator isample;


  // Resize all existing individuals
  // and set new elements to missing genotype (TF)
      
  if (par::SNP_major)
    {
      // Add space for new SNPs
      for (int i=0; i<locus2.size()-exist_cnt; i++)
	{
	  CSNP * newlocus = new CSNP;
	  newlocus->one.resize(n,true);
	  newlocus->two.resize(n,false);
	  SNP.push_back(newlocus);
	}      
      
    }
  else
    {
      // If using individual-major mode 
      for (int i=0; i<n; i++)
	{
	  sample[i]->one.resize(locus.size(),true);
	  sample[i]->two.resize(locus.size(),false);
	}
    }
  
  


  // An output file for diff mode
  ofstream MERD;
  if (par::merge_mode >=6)
    {
      string f = par::output_file_name+".diff";
      MERD.open(f.c_str(), ios::out);
      MERD << setw(20) << "SNP" << " " 
	   << setw(20) << "FID" << " " 
	   << setw(20) << "IID" << " " 
	   << setw(8) << "NEW" << " " 
	   << setw(8) << "OLD" << " " << "\n";
    }
  
  int new_person = 0;
  int old_person = 0;
  
  
  ///////////////////////////////////////
  // Read in FAM/BED file to new merge file
  
  bool fatal_error = false;


  // Initially, assume a binary trait
  par::qt = false;
  par::bt = true; 
  
  
  ifstream FAM;
  FAM.open(par::merge_famfile.c_str());
  FAM.clear();

  vector<bool> existing_person_list;
  vector<Individual*> sample2;

  c=0;
  while(!FAM.eof())
    {

      Individual * person = new Individual;

      // No comments allowed in BED/BIM/FAM files

      // First 6 obligatory fields
      string phenotype;

      FAM >> person->fid 
  	    >> person->iid 
	    >> person->pat
	    >> person->mat
	    >> person->sexcode
	    >> phenotype;
      
      // Skip last empty line that gets read
      if (person->fid=="") break;

      // Are we using 0/1 coding?
      if (par::coding01) 
	{
	  if ( phenotype == "1" ) 
	    phenotype = "2";      
	  else if ( phenotype == "0" )
	    phenotype = "1";
	  else 
	    phenotype = "0";
	}
            
      if (person->sexcode=="1")
	person->sex = true;  // male
      else if (person->sexcode=="2")
	person->sex = false; // female (default)
      else if (!par::ignore_missing_sex)
	{
	  person->missing = true;
	}

      
      // Have we already created this person?
      bool already_in = false;
      isample = msample.find(person->fid+"_"+person->iid);
      int indn = isample->second;
      if ( isample != msample.end() )
	{
	  already_in = true;
	  delete person;
	  person = sample[isample->second];
	  old_person++;
	}
      else
	new_person++;

      
      
      // Only look at phenotype if not already created
      if (!already_in)
	{
	  	

  	  //////////////////
	  // A non-founder?
	
	  person->founder = (person->pat == "0" && person->mat == "0") ? true : false;


	  /////////////////////////////////////////////////////
	  // Set missing status; test for quantitative traits?

	  if (phenotype == par::missing_phenotype)
	    person->missing = true;
	  else
	    {
	      person->phenotype = atof(phenotype.c_str());
	      
	      if (phenotype != "0" && 
		  phenotype != "1" && 
		  phenotype != "2" ) 
		{
		  par::qt = true;
		  par::bt = false; 
		}
	    }
	  	  
	}
      

      /////////////////////////////////////// 
      // Add necessary space for a new person 
      // Missing genotypes by default

      if (!already_in)      
	{
	  if (par::SNP_major)
	    {
	      // Add a new missing person to each SNP
	      vector<CSNP*>::iterator s = SNP.begin();
	      while ( s != SNP.end() ) 
		{
		  (*s)->one.push_back(true);
		  (*s)->two.push_back(false);
		  s++;
		}	      
	      // And set the individual number
	      indn = n + new_person - 1;
	    }
	  else
	    {
	      // Add all new SNPs to this person
	      person->one.resize(locus.size(),true);
	      person->two.resize(locus.size(),false);
	    }

	}


	// Record whether this individual is new or not
	existing_person_list.push_back(already_in);	      
      sample2.push_back(person);
 
      // Add individual to list, if need be
      if (!existing_person_list[c])
		{
	 	  sample.push_back(person);
		  msample.insert(make_pair( person->fid+"_"+person->iid , sample.size()-1 ) );
		}
	
      // Increase person counter
      c++;



   }

  if (!par::merge_list)
    {
      printLOG(int2str( existing_person_list.size() ) + " individuals merged from [ " + par::merge_famfile + " ] \n");
      printLOG("Of these, " + int2str(new_person) + " were new, " + int2str( old_person ) +
	       " were already in current data\n\n");
    }


   ////////////////////////////////////
   // Read genotype information, merge

  ifstream BIT;
  bool bfile_SNP_major = openBinaryFile( par::merge_bedfile, BIT ); 

  if (bfile_SNP_major != par::SNP_major)
	error("BED files must both be SNP-major or both individual-major for merging\n");

  if ( (!par::SNP_major) || (!bfile_SNP_major) ) 
	error("--bmerge not yet implemented for individual-major BED files -- please convert to SNP-major mode");



   ///////////////////////////
   // SNP-major mode
  
   if (bfile_SNP_major)
    {

      CSNP * snp;
      
      // Outer loop for SNPs
      int s=0;

      while (s<locus2.size()) // for all SNPs in second file
	{
	  
        // Look up SNP
//	  map<string,int>::iterator ilocus = mlocus.find(locus2[ s ]->name);
//	  int l2 = ilocus->second;
		
        // Inner loop for individuals
	     
	  vector<Individual*>::iterator person = sample2.begin();

	  while ( person != sample2.end() )
	    {
	      	      
	      char ch[1];
	      BIT.read(ch,1);
	      
	      bitset<8> b;
	      b = ch[0];	  

	      int c=0;

	      if (!BIT) 
		error("Problem with the BED file...has the FAM file been changed?\n");
	      
	      
	      while (c<7 && person != sample2.end() ) 
		{
		  
		  // Look up individual
		  map<string,int>::iterator isample = msample.find((*person)->fid+"_"+(*person)->iid);
        	  int indn;
		  if ( isample != msample.end() ) 
		    {
		      indn = isample->second;
		    }
         	  else
		    error("Internal error in --bmerge... should not happen...\n");

		  int k0 = (int)ordered[s].freq;
		  ilocus = mlocus.find(locus2[k0]->name);
		  int k = ilocus->second;
		  
		  bool s1 = b[c++];
		  bool s2 = b[c++];
		  
		  string one = locus2[ k0 ]->allele1;
		  string two = locus2[ k0 ]->allele2;
		  
		  if (s1 && s2) one=two=locus2[ k0 ]->allele2;
		  else if ( (!s1) && (!s2) ) one=two=locus2[ k0 ]->allele1;
		  else if ( s1 && !s2 ) one=two=par::missing_genotype;
		  		  
                  bool e = reconcileMerge( indn, k , 
 					   one, two, 
	 				   existing_person_list[indn], 
					   exists.find(locus2[k0]->name)->second,
					   MERD, misstrand);	
		  
		  if (e) fatal_error=true;
		  
		  // next person
		  person++;		  
		  indn++;
		}

	    }
	  
	  
	  // next SNP
	  s++;
	}
      
      // Set file mode
      par::SNP_major = true;
      
    }
   
   
   ////////////////////////////////////
   // Individual-major mode

//    else
//      {
       
//        // Outer loop for individuals
//        vector<Individual*>::iterator person = sample.begin();
//        while ( person != sample.end() )
// 	 {
	   
// 	   // Inner loop for SNPs
// 	   int s=0;
// 	   while (s<locus.size()) // for all SNPs
// 	     {
	       
// 	       char ch[1];
// 	       BIT.read(ch,1);
// 	       if (!BIT) error("Problem with the BED file... has the FAM file been changed?\n");
	       
// 	       bitset<8> b;
// 	       b = ch[0];	  
	       
// 	       int c=0;
	       
// 	       while (c<7 && s<locus.size())
// 		 {
// 		   (*person)->one[ s ] = b[c++];
// 		   (*person)->two[ s ] = b[c++];	      
// 		   s++;
// 		 }	 	  
// 	     }
	   
// 	   person++;
// 	 }
       
//        // Set file mode
//        par::SNP_major = false;
//      }
   
   
   
   
   
   //////////////
   // Close files
   
   FAM.clear();
   FAM.close();
   BIT.clear();
   BIT.close();
   
   
   
  ///////////////////////////////////////////////////////
  // Did we encounter any fatal errors from flipped SNPs? 
   
  if (fatal_error)
    {
      
      ofstream MSNP;
      string f = par::output_file_name+".missnp";
      MSNP.open(f.c_str(), ios::out);
      map<string,int>::iterator ilocus;
      
      for ( ilocus = misstrand.begin() ; ilocus != misstrand.end() ; ilocus++)
	{
	  MSNP << locus[ilocus->second]->chr << "\t"
	       << locus[ilocus->second]->name << "\n";
	}
      MSNP.close();
      cerr << "\nFound " << misstrand.size() << " SNPs that do not match in terms of strand\n";
      cerr << "(nb. flipped A/T and C/G SNPs will of course go undetected here)\n";
      cerr << "Writing identifiable strand problem SNPs to [ " << f << " ]\n";
      error("Stopping due to mis-matching SNPs -- check +/- strand?");
    }
  
  
  // If a binary trait, now make 0 missing also
  // i.e. if we never saw other than missing, 0, 1 or 2
  
  if (par::bt)
    for (int i=0; i<sample.size(); i++)
      if ( sample[i]->phenotype == 0 )
	sample[i]->missing = true;

  
  if (par::merge_mode >=6) 
    {
      printLOG("Results from diff ( merge mode " + int2str(par::merge_mode) + " ) written to [ " +
	       par::output_file_name + ".diff ]\n");
      MERD.close();

      shutdown();
    }
  
  


  // Phenotype statistics
  if (!par::merge_list)
    {
      int nm=0;
      for (int i=0;i<sample.size();i++)
	if(!sample[i]->missing) nm++;
      printLOG(int2str( nm ) + " individuals with nonmissing phenotypes\n");
      
      if (par::bt) printLOG( "Assuming a binary trait (1=unaff, 2=aff, 0=miss)\n");
      else printLOG("Assuming a quantitative trait\n");
      
      if (par::missing_phenotype!="0" && par::bt)
	printLOG( "Missing phenotype value is also " + par::missing_phenotype + "\n");
      else if (par::qt) 
	printLOG( "Missing phenotype value is " + par::missing_phenotype + "\n");
    }

}




