

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <cmath>
#include "options.h"
#include "plink.h"

#define EPS  0.00001
#define EPS2 0.0000001

using namespace std;

void Plink::preCalcSinglePoint()
{
  m1.resize(0);
  m2.resize(0);
  pos.resize(0);
    
  // All single markers
  for (int i=par::run_start;i<=par::run_end;i++)
    {
      m1.push_back(i);
      m2.push_back(i);
      pos.push_back(0);
    }

  // Final analysis is genome-wide phenotype on IBD 
  m1.push_back(-1);
  m2.push_back(-1);
  pos.push_back(0);


}

void Plink::preCalcMultiPoint()
{
  
  //////////////////////////////
  // Make multipoint marker map
  // Cover range pos1-fringe to pos(nl)+fringe

  // Either

  // SinglePoint (different function)
  // Uniform grid in cM space
  // Uniform grid in marker space (i.e. X positions between each marker pair).
  
  // Reset map

  m1.resize(0);
  m2.resize(0);
  pos.resize(0);
  
  // For each position on the cM map, determine the two 
  // flanking markers and the proportional distance between 
  // the two.

  // Uniform map in cM space
  if (par::inter_grid==0)
    {
      for (double p=locus[par::run_start]->pos - par::fringe;
	   p <= locus[par::run_end]->pos + par::fringe;
	   p += par::grid)
	{
	  
	  // find which marker this position comes before
	  
	  // before the first
	  if (p < locus[par::run_start]->pos)
	    { 
	      m1.push_back(-1);
	      m2.push_back(par::run_start);
	      double p2 = (p- (locus[par::run_start]->pos - par::fringe))
		/ par::fringe;
	      pos.push_back(p2); 
	    }
	  
	  for (int i=par::run_start; i<par::run_end; i++)
	    {	  
	      if (p >= locus[i]->pos && 
		  p < locus[i+1]->pos )
		{
		  double p2 = (p-locus[i]->pos)
		    / (locus[i+1]->pos - 
		       locus[i]->pos);
		  pos.push_back(p2);
		  m1.push_back(i);
		  m2.push_back(i+1);		  
		}
	    }
	  
	  // or after the last marker
	  if (p >= locus[par::run_end]->pos)
	    {
	      m1.push_back(par::run_end);
	      m2.push_back(-1);
	      double p2 = (p-locus[par::run_end]->pos)
		/ par::fringe;
	      pos.push_back(p2); 
	    }    
	}
    }
  else
    {
      // Uniform map in marker-space
      
      // before the first
      for (int j=0;j<par::inter_grid;j++)
	{
	  m1.push_back(-1);
	  m2.push_back(par::run_start);
	  pos.push_back((double)j/(double)par::inter_grid); 
	}
      
      // normal range
      for (int i=par::run_start; i<par::run_end; i++)
	{	  
	  for (int j=0;j<par::inter_grid;j++)
	    {
	      m1.push_back(i);
	      m2.push_back(i+1);
	      pos.push_back((double)j/(double)par::inter_grid); 
	    }
	}
      
      // after last marker
      for (int j=0;j<=par::inter_grid;j++)
	{
	  m1.push_back(par::run_end);
	  m2.push_back(-1);
	  pos.push_back((double)j/(double)par::inter_grid); 
	}
    }
  

  // Final analysis is genome-wide phenotype on IBD 
  m1.push_back(-1);
  m2.push_back(-1);
  pos.push_back(0);


  if (par::verbose)
    {
      cout << "Multipoint marker map information\n";
      for (int i=0; i<pos.size();i++)
	{
	  cout << i << "\t" 
	       << m1[i] << "\t" 
	       << m2[i] << "\t" 
	       << pos[i] << "\n";
	}
    }
 
}

vector<double> Plink::calcSinglePoint(vector<Z> & IBD, Z IBDg)
{
    vector<double> pihat;

    // Singlepoint individual loci
    for (int l=0; l<nl; l++)
    {	  
      
      //pihat.push_back( (IBD[l].z1*0.5+IBD[l].z2) - (IBDg.z1*0.5 + IBDg.z2)  );      
      pihat.push_back( (IBD[l].z1*0.5+IBD[l].z2) );      
      
      if (par::multi_output) 
	{
 	  cout << "S " 
	       << pairid << " " 
 	       << locus[m1[l]]->name << "\t"
	       << IBD[l].z0 << " "
	       << IBD[l].z1 << " "
	       << IBD[l].z2 << " => "
	       << IBD[l].z1*0.5+IBD[l].z2 << " "
	       << (IBDg.z1*0.5 + IBDg.z2) << "\n";	  
	}
    }

    // Final analysis is genome-wide IBD
    if (!par::done_global_pihat) pihat_G.push_back( (IBDg.z1*0.5 + IBDg.z2) );
    
    return pihat;
}


vector<double> Plink::calcMultiPoint(vector<Z> & IBD, Z IBDg)
{
  

  bool is0zero=false;
  bool is1zero=false;
  bool is2zero=false;
  
  if (IBDg.z0 < EPS) is0zero=true;
  if (IBDg.z1 < EPS) is1zero=true;
  if (IBDg.z2 < EPS) is2zero=true;
  

  
  // For m loci, m-1 cuts
  // 2(m-1) likelihood components
  

  ////////////////////////////////////////////
  // Build m-1 transition matrices
  // Input map distance from n to n+1 locus 
  //
  //      0        1       1        2  
  // 0 [ (1-T)^2   T(1-T)  T(1-T)   T^2     ]
  // 1 [ T(1-T)    (1-T)^2 T^2      T(1-T)  ]
  // 1 [ T(1-T)    T^2     (1-T)^2  T(1-T)  ]
  // 2 [ T^2       T(1-T)  T(1-T)   (1-T)^2 ]
  //
  //
  //      0         1               2  
  // 0 [ (1-T)^2    2T(1-T)         T^2     ]
  // 1 [ 2T(1-T)    2((1-T)^2+T^2)  2T(1-T) ]
  // 2 [ T^2        2T(1-T)         (1-T)^2 ]
  //
  //
  
  
  
  // Map distance x between two markers (Morgans) 
  //  m meioses ---  mx crossovers (cumulative)
  //
  // pihat = 1 / 2^m   
  // x = distance between two consecitive (Morgans) 
  // y = - ( ln pihat / ln 2 ) x 
  // T = ( 1 - e^{-2y} ) / 2
  
  
  // L = 1R Q1 T1 Q2 T2 .. T(M-1) Q(M) 1C 
  
  // 1R = 1x3 vector
  // 1C = 3x1 vector
  

  // L     R 
  // 1-2   4-5
  // 2-3   3-4
  // 3-4   2-3
  // 4-5   1-2

  
  
  ////////////////////////////////////////////
  // Q vector
  //
  // From ZL0, ZL1, ZL2,  [np][nl]
  //
  // [ z0   0     0  ] 
  // [  0   z1    0  ]
  // [  0   0     z2 ]
  

  // Populate this vector with pihat values
  vector<double> pihat;
  
  
  // Ignore if genome-wide pihat == 0 
  if ( (IBDg.z2 + IBDg.z1/2.0) < par::MIN_PIHAT ||
       (IBDg.z2 + IBDg.z1/2.0) > par::MAX_PIHAT )
    {
      cout << "THIS SHOULD NOT HAPPEN... \n";
      cerr << "THIS SHOULD NOT HAPPEN... \n";

      // return empty vector
      return pihat;
    }
  
  
  ///////////////////////////////////////////////
  // Now consider all other mutlipoint positions
  
  //double logPIover2 = log(IBDg.z2 + IBDg.z1/2.0 ) / log((double)2);
  
  // Working matrices
  vector<Z> left(nl);
  vector<Z> right(nl);
 
  // Q matrix diagonal elements
  double Q00, Q11, Q22;
  
  // Left conditional begins with first locus
  // on diagonal
  // (holds a Z structure)
  left[0].z0 = IBD[0].z0;
  left[0].z1 = IBD[0].z1;
  left[0].z2 = IBD[0].z2;
  
  left[0].z0 /= IBDg.z0;
  left[0].z1 /= IBDg.z1;
  left[0].z2 /= IBDg.z2;
  
  if (is0zero) left[0].z0 = 0;
  if (is1zero) left[0].z1 = 0;
  if (is2zero) left[0].z2 = 0;
  
  
  // Make conditional (sum to 1)
  double S = left[0].z0 + left[0].z1 + left[0].z2;
  left[0].z0 /= S;
  left[0].z1 /= S;
  left[0].z2 /= S;
  
  // FIRST Right conditional locus 
  right[nl-1].z0 = IBD[nl-1].z0;
  right[nl-1].z1 = IBD[nl-1].z1;
  right[nl-1].z2 = IBD[nl-1].z2;
  
  right[nl-1].z0 /= IBDg.z0;
  right[nl-1].z1 /= IBDg.z1;
  right[nl-1].z2 /= IBDg.z2;
  
  if (is0zero) right[nl-1].z0 = 0;
  if (is1zero) right[nl-1].z1 = 0;
  if (is2zero) right[nl-1].z2 = 0;
   
  
  // Make conditional (sum to 1)
  S = right[nl-1].z0 + right[nl-1].z1 + right[nl-1].z2;
  right[nl-1].z0 /= S;
  right[nl-1].z1 /= S;
  right[nl-1].z2 /= S;
  


  ///////////////////////
  // Left conditional 
  
  int l=1;

  for (int l2=par::run_start+1; l2<=par::run_end; l2++)
    {	  
      
      // populate T elements
      
      
      // 1 x 3 . 3x3 . 3x3 . ... 

      Z prev;

      prev.z0 = left[l-1].z0;
      prev.z1 = left[l-1].z1;
      prev.z2 = left[l-1].z2;

      buildT(locus[l2]->pos - locus[l2-1]->pos,  // Interval
	     IBDg);
	     
       
      
      // Q diagonal elemnts in IBD[l]
      // Left conditional [ R(n+1) * T * R(n) ] 
      left[l].z0 = ( ( prev.z0 * T00
		       + prev.z1 * T10
		       + prev.z2 * T20 ) * IBD[l].z0 );
      
      left[l].z1 = ( ( prev.z0 * T01
		       + prev.z1 * T11
		       + prev.z2 * T21 ) * IBD[l].z1 );
      
      left[l].z2 = ( ( prev.z0 * T02
		       + prev.z1 * T12
		       + prev.z2 * T22 ) * IBD[l].z2 );
      
      left[l].z0 /= IBDg.z0;
      left[l].z1 /= IBDg.z1;
      left[l].z2 /= IBDg.z2;
      
      if (is0zero) left[l].z0 = 0;
      if (is1zero) left[l].z1 = 0;
      if (is2zero) left[l].z2 = 0;
      
      // Make conditional (sum to 1)
      double S = left[l].z0 + left[l].z1 + left[l].z2;
      left[l].z0 /= S;
      left[l].z1 /= S;
      left[l].z2 /= S;
      
      // but check for this... investigate conditions when this occurs
      // e.g. mendel errors for parent-offspring seems most likely
      if (S < EPS) left[l] = IBDg;
      
      // Shift left
      l++;
      
    } // Next marker interval
  
  if (par::verbose)
    {
      cout << "LEFT CONDITIONAL\n";
      for (int i=0;i<left.size();i++)
	{
	  cout << left[i].z0 << "  " 
	       << left[i].z1 << "  " 
	       << left[i].z2 << "\n";
	}
    }
  
  
  
   ///////////////////////
   // Right conditional 
  
   l = nl-1-1;
 
   for (int l2=par::run_end-1; l2>=par::run_start; l2--)
     {	  
      
       buildT(locus[l2+1]->pos - locus[l2]->pos
 	     ,IBDg);

       // Right conditional [ R(n+1) * T * R(n) ] 
       right[l].z0 = ( ( right[l+1].z0 * T00
 			+ right[l+1].z1 * T10
 			+ right[l+1].z2 * T20 ) * IBD[l].z0 );
      
       right[l].z1 = ( ( right[l+1].z0 * T01
 			+ right[l+1].z1 * T11
 			+ right[l+1].z2 * T21 ) * IBD[l].z1 );
      
       right[l].z2 = ( ( right[l+1].z0 * T02
 			+ right[l+1].z1 * T12
 			+ right[l+1].z2 * T22 ) * IBD[l].z2 );
       
       right[l].z0 /= IBDg.z0;
       right[l].z1 /= IBDg.z1;
       right[l].z2 /= IBDg.z2;
       
       if (is0zero) right[l].z0 = 0;
       if (is1zero) right[l].z1 = 0;
       if (is2zero) right[l].z2 = 0;
       
       
       // Make conditional (sum to 1)
       double S = right[l].z0 + right[l].z1 + right[l].z2;
       right[l].z0 /= S;
       right[l].z1 /= S;
       right[l].z2 /= S;

       if (S < EPS) right[l] = IBDg;
     
       // Shift right
       l--;
      
     } // Next marker interval
  
  
   if (par::verbose)
     {
       cout << "RIGHT CONDITIONAL\n";
       for (int i=0;i<right.size();i++)
 	cout << right[i].z0 << "  " 
 	     << right[i].z1 << "  " 
 	     << right[i].z2 << "\n";
       cout <<"\n";
     }
  
  
   // Now we have LEFT and RIGHT tables of conditional 
   // probabilities -- rescan to get multipoint interpolation
  
  
  
   ///////////////////////////
   // Multipoint calculation
  
   for (int i=0; i<pos.size()-1; i++) // skip last position (gIBD)
     {
      
       double p1, p2;
     
       if (m1[i]==-1) 
 	p1 = locus[par::run_start]->pos - par::fringe;
       else p1 = locus[m1[i]]->pos;
      
       if (m2[i]==-1) 
 	p2 = locus[par::run_end]->pos + par::fringe;
       else p2 = locus[m2[i]]->pos;
      
       double d1 = pos[i] * (p2-p1);
       double d2 = (1 - pos[i]) * (p2-p1);
      
       // P0 = L * TL * Q0 * TR * R;
       // P1 = L * TL * Q1 * TR * R;
       // P2 = L * TL * Q2 * TR * R;

       //  L * TL

       // Left T matrix
       buildT(d1,IBDg);
       
       // Left & right conditional
       Z L;
       
       if (m1[i]==-1) { 
 	 L.z0 = L.z1 = L.z2 = 1;
       }
       else L = left[m1[i] - par::run_start];
       
       // * Q{0/1/2} [ Q 3x3 matrix -- just extracts elements ]
       
       Z M0;
       Z M1;
       Z M2;
       
       M0.z0 = ( L.z0 * T00
		 + L.z1 * T10
		 + L.z2 * T20 ) ;
       
       M1.z1 = ( L.z0 * T01
		 + L.z1 * T11
		 + L.z2 * T21 ) ;
       
       M2.z2 = ( L.z0 * T02
		 + L.z1 * T12
		 + L.z2 * T22 ) ;
       
       
       // * TR (transpose)
       buildT(d2,IBDg);
       
       // Finally, 1x3 . 3x1 = 11
       
       Z R;      
       
       if (m2[i]==-1) { 
	 R.z0 = R.z1 = R.z2 = 1;
       }
       else R = right[m2[i] - par::run_start];
       
       double P0 = M0.z0 * (( T00*R.z0 + T10*R.z1 + T20*R.z2)); 
       double P1 = M1.z1 * (( T01*R.z0 + T11*R.z1 + T21*R.z2)); 
       double P2 = M2.z2 * (( T02*R.z0 + T12*R.z1 + T22*R.z2)); 

       // Sum to 1 
       double S = P0 + P1 + P2;
       P0 /= S;
       P1 /= S;
       P2 /= S;
       
       // Bayes Theorem 
       S = P0*IBDg.z0 + P1*IBDg.z1 + P2*IBDg.z2;
       P0 = (P0*IBDg.z0) / S;
       P1 = (P1*IBDg.z1) / S;
       P2 = (P2*IBDg.z2) / S;
       
       if (par::multi_output)
	 {
	   double p1, p2;
	   string n1, n2;
	  
 	  if (m1[i]==-1) 
 	    {
 	      p1 = locus[par::run_start]->pos - par::fringe;
 	      n1 = "fringe";
 	    }
 	  else 
 	    {
 	      p1 = locus[m1[i]]->pos;
 	      n1 = locus[m1[i]]->name;
 	    }
	  
 	  if (m2[i]==-1) 
 	    {
 	      p2 = locus[par::run_end]->pos + par::fringe;
 	      n2 = "fringe";
 	    }
 	  else 
 	    {
 	      p2 = locus[m2[i]]->pos;
 	      n2 = locus[m2[i]]->name;
 	    }
	  
 	  cout << "M " 
	       << pairid << " " 
	       << p1 + pos[i] * (p2-p1) << " " 
 	       << n1 << " " << n2 << "\t"
	       << P0 << "  " 
 	       << P1 << "  " 
 	       << P2 << "  "
 	       << (P1*0.5+P2) << "  " 
 	       << (IBDg.z1*0.5 + IBDg.z2) << "\n";
 	}
      
       // Save pi-hat deviation from prior 
       // for this position/pair 
      
       //pihat.push_back( (P1*0.5+P2) - (IBDg.z1*0.5 + IBDg.z2) );	  
       pihat.push_back( (P1*0.5+P2)  );	  

     }
  

  // Final analysis is genome-wide IBD
  if (!par::done_global_pihat) pihat_G.push_back( (IBDg.z1*0.5 + IBDg.z2) );

   if (par::multi_output)
 	cout << "\n\n";
      
   return pihat;

}



void Plink::buildT(double cM, Z IBDg)
{  
  
  // Do not allow 0 cM distance between markers
  //  if (cM < EPS) cM = EPS;
  

  //////////////////////////
  // OLD 
  //  double t = (1-exp(-2*( -logPIover2 *cM )) )/2;
  //  double logPIover2 = log(IBDg.z2 + IBDg.z1/2.0 ) / log((double)2);  

  
  ///////////////////////////
  // Z0, Z1 and Z2

  double z0 = IBDg.z0;
  double z1 = IBDg.z1;
  double z2 = IBDg.z2;

  ///////////////////////////
  // Recombination fractions
  
  double G = cM;
  if (par::verbose)
    cout << "G = " << G << "\n";



  // For haploid genome pair

  // 00 ( 1 - (m-1)*G ) + (m-1)*G*(1-(1/(2^(m-1)-1)) ) 
  // 01 1 - 00
  // 10 (m-1)*G
  // 11 1 - (m-1)*G


  /////////////////////////////////////////
  // Calculate two IBD probabilities
  

  // P(haploid genome is IBD) = x1 and x2 

  // x1 * x2 = z2
  // (1-x1) * (1-x2) = z0

  // x^2 - x + 0.25 = 0 

  // Solution to this quadratic equation gives:

  // ax^2 + bx + c = 0

  // a = -1
  // b = 1-z0+z2
  // c = -z2

  double mA, mB;
  bool z2zero=false;

  if ( z2 < EPS2) z2zero = true;

  if (z2zero)
    {
      mA = 1 - log(z1) / log(2.0);
      mB=0;
    }
  else
    {
      double b = 1 - z0 + z2;

      // After 'nudging' IBD probabilies, this should
      // always be positive -- but allow for rounding
      // errors (-0) with the fabs option.
      double b2 = fabs( b*b - 4*z2 );
      double x = sqrt( b2 );
      
      //  =(-E13-SQRT(E13*E13-4*E12*E14))/(2*E12)
      
      double t1 = (-b + x ) / -2;
      double t2 = (-b - x ) / -2;

      mA = 1 - log(t1) / log(2.0);
      mB = 1 - log(t2) / log(2.0);

      if (par::verbose) 
	{
	  cout << "  b = " << b ;
	  cout << "  b^2 = " << b*b;
	  cout << "  4z2 = " << 4*z2;
	  cout << "  b^2-4z2 = " << b*b - 4*z2;
	  cout << "  x = " << x ;
	  cout << "  t1, t2 = " << t1 << " " << t2 << "\n";
	}
    }

  if (par::verbose)
    {
      cout << "WORKING...\n"
	   << "-------------------------------------------\n\n";
      
      cout << "Z0, Z1, Z2 = " << z0 << " " << z1 << " " << z2 << "\n";
      
      cout << "mA, mB = " << mA << " " << mB << "\n";
      cout << "\n";
    }


  /////////////////////////////////////////
  // Build 2x2 haploid transition matrices

  //                  m
  //                 ------------
  //                 0        1
  //                 ------------
  //
  //  m-1 |  0  |    X        t
  //      |  1  |    1-X      1-t

  // 
  // where t = mA * G = recombination fraction theta
  //      mA = number of meioses separating the haploid genomes
  //       G = genetic distance in Morgans
  // 
  //  and  X = ( 1 - t ) + t( 1 - 1/(2^(mA-1)-1) )
  // 
  //   i.e. of staying IBD 0 (either no recombination, or 
  //        recombination but to chromosome other than self or partner

  
  // Taking these 2 possibilities together, the probability that A and
  // B and IBD at V, given that they are IBD at U, is (1-c)^m+
  // c^2(1-c)^(m-2). This can be rewritten as
  // (1-c)^(m-2)(c^2+(1-c)^2).  The probability that A and B are not
  // IBD at V, given that they are IBD at U, is therefore
  // 1-(1-c)^(m-2)(c^2+(1-c)^2).

  // We can now find the probability that A and B are IBD at V, given
  // that they are not IBD at U, by Bayes' theorem as follows:

  // P(IBD at V | Not IBD at U) = P(Not IBD at U | IBD at V) P (IBD at V) 
  // / P(Not IBD at U)

  // This simplifies to (1-(1-c)^(m-2)(c^2+(1-c)^2))/(2^(m-1)-1)

  // Finally, the probability at A and B are not IBD, given that they are 
  // not IBD at U, is simply 1 - (1-(1-c)^(m-2)(c^2+(1-c)^2))/(2^(m-1)-1)

  // Note that for small values of c, (1-c)^m+ c^2(1-c)^(m-2) is 
  // approximately equally to 1-mc.

  
  
  double A01 = (1- pow(1-G,mA-2) * (G*G+(1-G)*(1-G)))/(pow(2,mA-1)-1);
  double A00 = 1 - A01;
  double A11 = pow((1-G),(mA-2)) * (G*G+(1-G)*(1-G) );
  double A10 = 1 - A11;
  
  double B00;
  double B01;
  double B10;
  double B11;


  if (z2zero) 
  {
    B00 = B11 = 1;
    B01 = B10 = 0;
  }
  else
    {
      B01 = (1- pow(1-G,mB-2) * (G*G+(1-G)*(1-G)))/(pow(2,mB-1)-1);
      B00 = 1 - B01;
      B11 = pow((1-G),(mB-2)) * (G*G+(1-G)*(1-G) );
      B10 = 1 - B11;
    }
 

  if (par::verbose)
    {
      cout << "A[i,j] = \n";
      cout << "\t" << A00 << "\t" << A01 << "\n"
	   << "\t" << A10 << "\t" << A11 << "\n\n";
      
      cout << "B[i,j] = \n";
      cout << "\t" << B00 << "\t" << B01 << "\n"
	   << "\t" << B10 << "\t" << B11 << "\n\n";
    }

  ///////////////////////////
  // Build transition matrix
  
  T00 = A00*B00;              
  T01 = A00*B01 + A01*B00;       
  T02 = A01*B01;
  
  T10 = A00*B10 + A10*B00; 
  T11 = A00*B11 + A01*B10 + A10*B01 + A11*B00;
  T12 = A01*B11 + A11*B01;
  
  T10 /= 2;
  T11 /= 2;
  T12 /= 2;

  T20 = A10*B10; 
  T21 = A10*B11 + A11*B10;
  T22 = A11*B11;
  

  if (par::verbose)
    {
      cout << "cM = " << cM <<  "\n";
      cout << "transition matrix\n"
	   << T00 << "\t" << T01 << "\t" << T02 << "\n"
	   << T10 << "\t" << T11 << "\t" << T12 << "\n"
	   << T20 << "\t" << T21 << "\t" << T22 << "\n";
    }

}

