

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <vector>
#include <map>
#include <list>
#include <time.h>
#include <stdlib.h>
#include <algorithm>

#include "plink.h"
#include "options.h"
#include "helper.h"
#include "crandom.h"
#include "perm.h"
#include "sets.h"
#include "linear.h"

// Global variables
using namespace std;
ofstream     LOG;
string PVERSION;
string PDATE;

int main(int argc, char* argv[]) 
{


  /////////////////////////
  // Setup, display title

  CRandom::srand(time(0));
  cout.setf(ios::fixed);
  cout.precision(4);

  set_new_handler(NoMem);

  PVERSION = "0.99l";       // 5 chars
  PDATE    = "27/Jul/2006"; // 11 chars


  //////////////////
  // The major class

  Plink P;


  //////////////////////////
  // Command line arguments

  CArgs a(argc,argv);
  setOptions(a);


  //////////////////////////
  // Start logging, title

  LOG.open(string(par::output_file_name + ".log").c_str());
  P.printLOG("\n"
	     "@----------------------------------------------------------@\n"
	     "|         PLINK!       |    v"+PVERSION+"     |   "+PDATE+"     |\n"
	     "|----------------------------------------------------------|\n"
	     "|  (C) 2006 Shaun Purcell, GNU General Public License, v2  |\n"
	     "|----------------------------------------------------------|\n"
	     "|       http://pngu.mgh.harvard.edu/purcell/plink/         |\n"
	     "@----------------------------------------------------------@\n"
	     "\n");


  /////////////////////
  // Permutation class

  Perm perm(P);



  ////////////////
  // Check version 
  
  if (par::web_check)
    P.webcheck();
  else
    P.printLOG("Skipping web check... [ --noweb ] \n");
  

  P.printLOG("Writing this text to log file [ "+
	     par::output_file_name + ".log ]\n");
    
  time_t curr=time(0);
  string tdstamp = ctime(&curr);
  P.printLOG("Analysis started: " + tdstamp +"\n");


  /////////////////////////////////////
  // Validate and record all arguments

  a.check_unused_options(P);

  
  //////////////////////////
  // Some basic definitions

  if (par::species_dog) defineDogChromosomes();
  else defineHumanChromosomes();


  /////////////////////////
  // Input files

  // Binary or ASCII format	  
  if (par::read_bitfile) P.readBinData();
  else P.readData();  

  // Set number of individuals
  P.n = P.sample.size();
  
  // Set number of pairs
  P.np = (int)((double)(P.n*(P.n-1))/(double)2);
  
  // Total number of all (test+background) loci
  P.nl_all = P.locus.size();

  // Number of permutations to store
  P.maxr2.resize(par::replicates);


  /////////////////////////////////////
  // Merge with a secondary data file 
  // Standard (non-list) mode

  if (par::merge_data && !par::merge_list)
    {
      
      if (par::merge_binary)
	P.mergeBinaryData();
      else
	P.mergeData();
      
      // Reset number of individuals
      P.n = P.sample.size();
      
      // Set number of pairs
      P.np = (int)((double)(P.n*(P.n-1))/(double)2);
      
      // Total number of all (test+background) loci
      P.nl_all = P.locus.size();
    }
  
  
  /////////////////////////////////////
  // Merge with a secondary data file 
  // List mode
  
  if (par::merge_list)
    P.mergeList();
      
  

  //////////////////////////////////////////  
  // A different phenotype file specified?

  if (par::pheno_file) P.readPhenoFile();


  //////////////////////////////// 
  // A covariate file specified?

  if (par::covar_file) 
    {
      if (!P.readCovariateFile()) 
	error("Problem reading the specified covariate from the covariate file");
    }
  

  ////////////////////////////////
  // Remove any individuals with 
  // missing phenotypes
  
  if (!par::ignore_phenotypes)
    removeMissingPhenotypes(P);


  //////////////////////////////////
  // Binary affection status coding

  if (par::bt)
    affCoding(P);
        
  
  //////////////////////////////////////////////////////////
  // Output a specific set of SNPs (--extract or --exclude)

  if (par::dump_set)
    {
      P.extractExcludeSet(false);      
    }
  else if (par::exclude_set)
    {
      P.extractExcludeSet(true);
    }  

  /////////////////////////////////////////////////////////////
  // Output a specific set of individuals (--remove or --keep)

  if (par::remove_indiv)
    {
      P.removeIndividuals(par::remove_keep_indiv);
    }


  /////////////////////////////////
  // Flip DNA strand for any SNPs? 

  if (par::flip_strand)
    {
      P.flipStrand();
    }

 
 
  //////////////////////////////////////
  // Asign cluster solution from file 

  if (par::include_cluster_from_file)
    {
      P.printLOG("Reading clusters from [ " + 
		 par::include_cluster_filename+" ]\n");
      if (!P.readClusterFile())
	error("Problem reading from [ "+par::include_cluster_filename+" ]");
    }
  

  /////////////////////////////////////////////
  // Assign cluster "solution" to FAM_ID column

  if (par::sol_family)
    {
      
      P.printLOG("Setting to permute within family ID only\n");
      vector<string> famlist;
      P.kname.resize(0);
      for (int i=0; i<P.n; i++)
	{
	  Individual * person = P.sample[i];
	  bool match = false;
	  for (unsigned int j=0; j<famlist.size(); j++)
	    if (person->fid == famlist[j]) 
	      {
  		match=true;
		person->sol=j;
	      }	  
	  if (!match)
	    {
	      famlist.push_back(person->fid);
	      person->sol=famlist.size()-1;
	      P.kname.push_back(person->fid);
	    }
	}

      // Set number of clusters/families
      P.nk = famlist.size();     
    }


  //////////////////////////////////
  // Determine formats for printing

  P.prettyPrintLengths();


  /////////////////////////////////////
  // Calcaulte allele freqs (or --FIX)
  
  P.printLOG("Before frequency and genotyping pruning, there are "+int2str(P.nl_all)+" SNPs\n");

  if (!par::FIXED)
    {
      if (par::SNP_major)
	P.filterSNPs_SNP();
      else
	P.filterSNPs_individual();
    }
  else
    for (int i=0; i<P.nl_all;i++)
      {
	if (P.locus[i]->allele1=="1")
	  P.locus[i]->freq = par::FIX_p;
	else 
	  P.locus[i]->freq = 1-par::FIX_p;
      }

  P.printLOG("After frequency and genotyping pruning, there are "+int2str(P.nl_all)+" SNPs\n");



  ///////////////////////
  // Create family units?
  
  if (par::MENDEL_test || 
	par::MENDEL_report || 
	par::TDT_test ||
	par::QTDT_test &&
	!par::built_families)
    {
      map<string,Individual*> fnd;
      map<Individual*,int> idmap;
      P.linkRelateds(idmap, fnd);
      P.parseTrios();      
      par::built_families = true;
    }



  ///////////////////////////////////////////////////
  // Output a recoded file (optional 1/2 SNP coding)
  // or flip strand of every SNP

  if (par::recode) 
    {
      P.display_recoded_PEDFILE();
      shutdown();
    }

  if (par::recode_12) 
    {
      P.display_recoded_PEDFILE_12();
      shutdown();
    }

  if (par::recode_AD) 
    {
      P.display_recoded_PEDFILE_AD();
      shutdown();
    }


  /////////////////////////////
  // Output a recoded BIT file 

  if (par::write_bitfile) 
    {
      P.write_BITFILE();
      shutdown();
    }



  //////////////////////////////////////////
  // Warn if not enough markers in analysis

  if (par::plink || par::cluster || par::genome_output || par::inbreeding)
	{
          if (P.nl_all < 1000) 
	    P.printLOG("\n **Warning** this analysis option requires whole-genome level data\n"
		       "             to give accurate results ( probably >10K SNPs ) \n\n");
      }



  /////////////////////////////////////
  // Create tags, haplotype-based tests

  if (par::phase_snps)
    {

      // In case families are included
      if (par::has_nonfounders && !par::built_families)
	{
	  map<string,Individual*> fnd;
	  map<Individual*,int> idmap;
	  P.linkRelateds(idmap, fnd);
	  P.parseTrios();
	  P.checkMendel();      
	  par::built_families = true;
	}

      P.readTagFile();
      shutdown();
    }



  /////////////////////////////////
  // Perform a cluster analysis 

  if (par::cluster) 
	{ 
	  P.buildCluster();
	  shutdown();
	}
  

  

  /////////////////////////
  //  Pre-Analysis 
  
  // Calculate SD or CP for phenotype
  if (!par::ignore_phenotypes) P.preCalcPhenotypes();

  // Precalculate frequency-averaged P(IBD|IBS) table
  if (par::plink || par::genome_output) 
    P.preCalcGenomeIBD();

  
  //////////////////////////////
  // Genome-wide output only
 
  if (par::genome_output)
    P.displayGenomeWideInfo();
      
  
  
  //////////////////////////////
  // Genome-wide inbreeding output only
 
  if (par::inbreeding)
    {
      
      if (par::SNP_major) P.SNP2Ind();

      ofstream HET;
      string f = par::output_file_name + ".het";
      HET.open(f.c_str(),ios::out);
      HET.precision(4);
        
      P.printLOG("Writing individual heterozygosity information to [ "+f+" ] \n");
      HET << setw(par::pp_maxfid) << "FID" << " "
	  << setw(par::pp_maxiid) << "IID" << " "
	  << setw(12) << "O(HOM)" << " " 
	  << setw(12) << "E(HOM)" << " " 
	  << setw(12) << "N(NM)" << " " 
	  << setw(12) << "F" << "\n";
      
      for (int i1=0; i1<P.n; i1++)
	P.calcInbreeding(P.sample[i1],0,P.nl_all-1,HET);
      
      HET.close();
      
  }


  //////////////////////////////
  // Local homozygosity estimation
  
  if (par::local_homozygosity)
    {
  
      if (par::SNP_major) P.SNP2Ind();
      
      ofstream HET;
      string f = par::output_file_name + ".lhet";
      HET.open(f.c_str(),ios::out);
      
      P.printLOG("Writing individual heterozygosity information to [ "+f+" ] \n");
      
      for (int i1=0; i1<P.n; i1++)
	{
	  cout << i1+1 << " of " << P.n << " individuals      \r";
	  
	  HET << P.sample[i1]->fid << "\t" << P.sample[i1]->iid << "\t";

	  for (int l=0; 
	       l<P.nl_all-par::local_homozygosity_window; 
	       l+=par::local_homozygosity_step)
	    {
	      // if window does not span chromosomes...
	      if (P.locus[l]->chr == 
		  P.locus[l+(int)par::local_homozygosity_window]->chr)
		P.calcInbreeding(P.sample[i1], l , 
				 l+(int)par::local_homozygosity_window, HET );
	      
	    }
	  HET << "\n";
	}
      
      HET.close();
     
      shutdown();
    }

  ///////////////////////
  // Runs of homozygosity
  
  if (par::homo_run)
    {
      
      if (par::SNP_major) P.SNP2Ind();

      ofstream HOM;
      string f = par::output_file_name + ".hom";
      HOM.open(f.c_str(),ios::out);
      HOM.precision(4);
      
      P.printLOG("Writing homozygosity-run information to [ "+f+" ] \n");
      string msg = "Run defined as "+int2str(par::homo_run_length);
      if (par::homo_run_length_kb) msg += " kb\n";
      else msg += " SNPs\n";      
      msg += "Allowing"+int2str(par::homo_run_het)+" heterozygotes per run\n";
      P.printLOG(msg);
 
   
      HOM << setw(par::pp_maxfid) << "FID" << " "
	  << setw(par::pp_maxiid) << "IID" << " "
	  << setw(4) << "CHR" << " "
	  << setw(par::pp_maxsnp) << "SNP1" << " "
	  << setw(par::pp_maxsnp) << "SNP2" << " "
	  << setw(12) << "POS1" << " "
	  << setw(12) << "POS2" << " "
	  << setw(10) << "KB" << " "
	  << setw(10) << "NSNP" << " "
	  << setw(4) << "NHET" << " "
	  << setw(4) << "NMISS" << "\n";


      for (int i1=0; i1<P.n; i1++)
	{
	  cout << i1+1 << " of " << P.n << " individuals      \r";
	  P.findHomoRuns(P.sample[i1], HOM);
	}
      
      cout << "\n\n";
      
      HOM.close();

      shutdown();
  }


  ///////////////////////////////////
  // Runs of missingness (deletions)
  
  if (par::miss_run)
    {

      if (par::SNP_major) P.SNP2Ind();

      ofstream RUN;
      string f = par::output_file_name + ".rum";
      RUN.open(f.c_str(),ios::out);
      
      P.printLOG("Writing run-of-missings information to [ "+f+" ] \n");

      string msg = "Run defined as " + int2str(par::miss_run_length);
      if (par::miss_run_length_kb) msg += " kb\n";
      else msg += " SNPs\n";
      P.printLOG(msg);

      stringstream s2;
      s2 << "With at least " << par::miss_run_level << " missingness\n";
      P.printLOG(s2.str());

      for (int i1=0; i1<P.n; i1++)
	{
	  cout << i1+1 << " of " << P.n << " individuals      \r";
	  P.findMissRuns(P.sample[i1],RUN);
	}
      cout << "\n\n";
      
      RUN.close();

      shutdown();
  }
  
  

  //////////////////////////////
  // Genome-wide association test

  if (par::assoc_test)
    {
      if (par::CMH_test_1 || par::CMH_test_2)
	  P.calcMH();
      else if (par::OR_homog_test)
	  P.calcHomog();
      else if (par::QTDT_test)
	{
	  // Force a Mendel error check
	  if (! (par::MENDEL_report || par::MENDEL_test) )
	    P.checkMendel();
	  P.perm_testQTDT(perm);
	}
      else if (par::boot)
	  P.calcAssociationWithBootstrap();
      else
	  P.calcAssociationWithPermutation(perm);
      shutdown();
    }

  //////////////////////////////
  // QTL interaction test

  if (par::assoc_gxe)
    {
      P.perm_testGXE2(perm);
      shutdown();
    }

  //////////////////////////////
  // QTL interaction test

  if (par::hotel)
    {
      P.perm_testHotel(perm);
      shutdown();
    }

  ///////////////////////////////////
  // Test difference in missing rates

  if (par::test_missing)
    {	
      P.calcAssociationWithPermutation(perm);
      shutdown();
    }
  

  //////////////////////////////////
  // MENDEL error checks?
  
  if (par::MENDEL_report || par::MENDEL_test)
    P.checkMendel();
  
  
  //////////////////////////////////
  // Genome-wide family-based (TDT)
  // and Parent-of-origin analysis

  if (par::TDT_test)
    {

      // Force a Mendel error check, if we have not 
      // already

      if (! (par::MENDEL_report || par::MENDEL_test) )
	P.checkMendel();

      // Either basic TDT or Parent-Of-Origin analysis

      if (par::parent_of_origin)
	P.perm_testTDT_POO(perm);
      else	  
	P.perm_testTDT(perm);
      shutdown();
    }
  

  //////////////////////////////
  // Genome-wide IBS sharing test

  if (par::ibs_sharing_test)
    {
      P.perm_sharingIBSTest(perm);
      shutdown();
    }


  //////////////////////////////
  // Gene-based test of epistasis

  if (par::epi_genebased)
    {
      P.driverSCREEPI();
      shutdown();
    }


  //////////////////////////////
  // Genome-wide epistasis tests

  if (par::epistasis)
    {
      P.calcEpistasis();      
      shutdown();
    }



  //////////////////////////////
  // LD statistics

  if (par::disp_r1 || par::disp_r2)
    {
      P.calcLDStatistics();
      shutdown();
    }


  ///////////////////
  // LD-based pruning
  
  if (par::prune_ld)
    {  
      P.pruneLD();
      shutdown();
    }

  
  
  // Stop now, unless a plink analysis is specified
  if (!par::plink) 
    shutdown();


  
  //////////////////
  // PLINK analysis
  
  if (par::SNP_major) 
    P.SNP2Ind();

  
  //////////////////////////////
  // Pair inclusion/exclusion

  int c=0; // # of informative pairs

  // Read or calculate informative pairs?
  if (par::inc_read) 
    {
      // Returns number of informative pairs
      c = P.readInformative();
    }
  else
    {
      c = P.calcInformative();
    }
  

  
  ///////////////////////////////////
  // Save pairs to be included? i.e. 
  // after removing all pairs for 

  // a) low IBD
  // b) not being an affected pair
  // c) being a concordant unaffected pair

  if (par::inc_write) P.writeInformative();

  /////////////////////////////////////////////////////////////////
  // Get and display information on chromosomal range to be tested
  
  if (par::bins)
    P.printLOG("Using bins analysis mode\n");
  else if (par::singlepoint)
    P.printLOG("Using singlepoint analysis mode\n");
  else if (par::inter_grid>0)
    {
      stringstream s2;
      s2 << "Using multipoint analysis: step = " 
	 << par::inter_grid 
	 << " and fringe = " 
	 << par::fringe << " cM\n";
      P.printLOG(s2.str());
    }
  else
    {
      stringstream s2;
      s2 << "Using multipoint analysis: grid = " 
	 << par::grid 
	 << " and fringe = " 
	 << par::fringe << " cM\n";
      P.printLOG(s2.str());
    }

  vector<int> chrs;
  if (par::run_chr==0) {
    vector<int> r = getChromosomeRange(P);
    if (r[0]<1) r[0]=1;
    if (r[1]>22) r[1]=22;
    P.printLOG("\nScanning from chromosome "+
	       int2str(r[0])+" to "+
	       int2str(r[1])+"\n\n");
    for (int i=r[0];i<=r[1];i++) 
      if ( ( !par::chr_haploid[i] ) && 
	   ( !par::chr_sex[i] ) )
	chrs.push_back(i);
  } else chrs.push_back(par::run_chr);
  

  // Blank list of segments
  P.segment.resize(0);

  ofstream SEG;
  if (par::segment_output)
    {
      string f = par::output_file_name + ".segment";
      SEG.open(f.c_str(),ios::out);
      P.printLOG("Writing IBD-segment information to [ "+f+" ]\n");

      f = par::output_file_name + ".segment.summary";
      P.printLOG("Writing IBD-segment summary to [ "+f+" ]\n\n");
      
    }



  ///////////////////////////
  // Consider each chromosome

  for (int i=0;i<chrs.size();i++)
    {
      
      // Reset main variables
      P.phenotype.resize(0);
      P.pihat.resize(0);
      P.Zlocus.resize(0);
      P.m_pihat.resize(0);
      P.v_pihat.resize(0);
      P.pair1.resize(0);
      P.pair2.resize(0);
      
      // Set chromosome
      par::run_chr = chrs[i];
      
      // Find scan range
      P.setMarkerRange();
      
      // Total number of all (test+background) loci
      P.nl_all = P.locus.size();
      
      // Number of loci for test loci
      P.nl = par::run_end - par::run_start + 1; 
      P.printLOG(int2str(P.nl)+" markers in this scan\n");
      
      // Set up single/multipoint marker map
      if (par::bins) 
	P.preCalcBins();
      else if (!par::singlepoint)
	P.preCalcMultiPoint();
      else
	P.preCalcSinglePoint();

           
      /////////////////////////////////
      //  For each pair of individuals
      //  calculate mutlipoint pihats
      
      // reset counter
      int c1=0;
      int c2=0;
      
      for (int i1=0; i1<P.n-1; i1++)
	for (int i2=i1+1; i2<P.n; i2++)
	  {
	    
	    if (!P.skip_pair[c1++])
	      {
		Individual * p1 = P.sample[i1];
		Individual * p2 = P.sample[i2];
		
		/////////////////////////////////
		//  1. Calculate IBD(g) | IBS(g) 
		
		// Note: could be outside of chromosome loop, 
		// but this way there is no need to store the values
		
		Z IBDg = P.saved_IBDg[c2];

		cout << "IBD calculation: " 
		     << ++c2
		     << " of " 
		     << c 
		     << "                  \r";
		cout.flush();

		
		/////////////////////////////////
		//  2. Calculate IBD(l) - IBD(g)
		
		vector<Z> IBDl = P.calcLocusIBD(p1,p2,IBDg);
		
		
		/////////////////////////////////
		//  3. Multipoint calculation
				
		P.pairid = itoa((int)p1->phenotype,10)+" "+itoa((int)p2->phenotype,10)+" ";
		P.pairid += itoa((int)c2,10) + " ";
		P.pairid += p1->fid+"_"+p1->iid+"_"+" x ";
		P.pairid += p2->fid+"_"+p2->iid+"_";
		
		vector<double> p;
		
		//////////////////////////////////
		// Perform either using
		//     Marker bins (experimental)
		//     Singlepoint analysis
		//     Multipoint analysis

		if (par::bins) p = P.calcBins(IBDl,IBDg);
		else if (!par::singlepoint) p = P.calcMultiPoint(IBDl,IBDg);
		else p = P.calcSinglePoint(IBDl,IBDg);
		

		////////////////////////////////
		//   3b. Verbose output:
		//       genotypes for each pair
		
		if (par::gmulti_output)
		  {
		    for (int l=par::run_start; l<=par::run_end; l++)
		      {
			cout << "G " << P.pairid << " " 
			     << P.locus[l]->name << " ";
			if ((!p1->one[l]) && (!p1->two[l])) 
			  cout << P.locus[l]->allele1 << " " 
			       << P.locus[l]->allele1 << "   ";
			else if ((!p1->one[l]) && p1->two[l]) 
			  cout << P.locus[l]->allele1 << " " 
			       << P.locus[l]->allele2 << "   ";
			else if (p1->one[l] && p1->two[l]) 
			  cout << P.locus[l]->allele2 << " " 
			       << P.locus[l]->allele2 << "   ";
			else if (p1->one[l] && (!p1->two[l])) 
			  cout << "0 0   ";
			
			if ((!p2->one[l]) && (!p2->two[l])) 
			  cout << P.locus[l]->allele1 << " " 
			       << P.locus[l]->allele1 << "   ";
			else if ((!p2->one[l]) && p2->two[l]) 
			  cout << P.locus[l]->allele1 << " " 
			       << P.locus[l]->allele2 << "   ";
			else if (p2->one[l] && p2->two[l]) 
			  cout << P.locus[l]->allele2 << " " 
			       << P.locus[l]->allele2 << "   ";
			else if (p2->one[l] && (!p2->two[l])) 
			  cout << "0 0   ";
			cout << "\n";
		      }
		  }
		
		if (par::multi_output)
		  cout << "\n\n";
		
		

		///////////////////////////////
		//  4. Scan for segments of IBD
		
		bool inseg = false;
		
		Segment s;
		s.p1 = p1;
		s.p2 = p2;
		int npos = p.size();

		for (int l=0; l<npos; l++)
		  {
		    
		    ///////////////////////////
		    // Start of a new segment?

		    if ( (!inseg) && p[l] >= par::segment_threshold_start)
		      {
			inseg = true;
			
			if (P.m1[l] != -1)
			  s.start = P.locus[P.m1[l]];
			else
			  s.start = P.locus[par::run_start];
		      }
		    

		    ///////////////////////////
		    // End of existing segment?
		    		    
		    if ( inseg && 
			 ( p[l] < par::segment_threshold_finish || l == npos-1 ) )
		      {
			inseg = false;;
			
			if (P.m2[l] != -1)
			  s.finish = P.locus[P.m2[l]];
			else
			  s.finish = P.locus[par::run_end];

			// Add segment to list
			P.segment.push_back(s);

			// Display?
			if (par::segment_output)
			  {
			    SEG << setw(10) << s.p1->fid << " "
				<< setw(10) << s.p1->iid << " "
				<< setw(10) << s.p2->fid << " "
				<< setw(10) << s.p2->iid << " ";

			    if (par::bt)
			      {
				if ( (!p1->aff) && (!p2->aff) ) 
				  SEG << setw(4) << "-1" << " ";
				else if ( p1->aff && p2->aff )
				  SEG << setw(4) << "1" << " ";
				else if ((!p1->aff) && p2->aff)
				  SEG << setw(4) << "0" << " ";
				else if (p1->aff && !p2->aff)
				  SEG << setw(4) << "0" << " ";
				else
				  SEG << setw(4) << "NA" << " ";
			      }
			    			    
			    SEG << setw(4) << par::run_chr << " "
				<< setw(10) << s.start->bp << " "
				<< setw(10) << s.finish->bp << " "
				<< setw(10) << (double)(s.finish->bp - s.start->bp)/(double)1000 
				<< "\n";
			    
			  }
			
		      }
		    
		    
		  }
		

		/////////////////////////////
		//  5. Add to list 

		P.pihat.push_back(p);
		
		// And (A,B) pair to list
		P.pair1.push_back(i1);
		P.pair2.push_back(i2);
		
	      }
  	    	    
	  }
      
     
      cout << "\n";


      /////////////////////////////////
      //  Make list of unique individuals
      
      // copy first set of individuals
      P.in_anal = P.pair1;
      for (unsigned int ind=0; ind<P.pair2.size(); ind++) 
	P.in_anal.push_back(P.pair2[ind]);
      sort(P.in_anal.begin(),P.in_anal.end());
      vector<int>::iterator new_end=
	unique(P.in_anal.begin(), P.in_anal.end());
      // delete all elements past new_end 
      P.in_anal.erase(new_end, P.in_anal.end());
      
      P.printLOG(int2str(P.in_anal.size())+" unique, informative individuals in analysis\n");
      
      if ( P.in_anal.size() == 0 )
	{
	  error("No individuals left in analysis: halting");
	}
	   
      /////////////////////////////////
      //  Verbose output: summarise IBD

	 if (par::segment_output)
	   {
	     P.summaryIBDsegments();
	   }
	 else
	   if (par::summary_ibd_output)
	     P.summaryIBD();

      
      /////////////////////////////////
      //  Regression analysis
            
	 P.printLOG("Analysis of "+int2str(P.pair1.size()) +
		    " informative pairs at " +int2str(P.pihat[0].size())+
		    " positions\n");
	 
	 P.calcRegression(i);
	 
	 
      /////////////////////////////////
      //  Next chromosome
      
      par::done_global_pihat = true;
     
      cout << "\n";
      
    }
      


  if (par::segment_output)
    SEG.close();
  
  
  ////////////////////////////////
  //  Output genome-wide p-values

  if (chrs.size()>=1 && (!par::ignore_phenotypes))
     P.displayGenomePV();
  
  

  

  ////////////////////////////////
  //  Done
  
  shutdown();

}


