

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <map>
#include <cmath>


#include "plink.h"
#include "helper.h"
#include "options.h"
#include "linear.h"
#include "perm.h"
#include "crandom.h"


void Plink::perm_testQTDT(Perm & perm)
{

  // Use individual-major coding
  if (par::SNP_major) 
    SNP2Ind();
  
  // Set up some necessary things
  
  vector<int> pbetween(family.size());
  vector<bool> pwithin(family.size(),false);
  for (int i=0; i < family.size(); i++)
    pbetween[i] = i;
  
  
  ///////////////
  // Output files

  string f = par::output_file_name + ".qfam";
  if (par::QFAM_within1) f += ".within";
  else if (par::QFAM_within2) f += ".parents";
  else if (par::QFAM_total) f += ".total";

  printLOG("\nWriting QFAM results to [ "+f+" ]\n");
  ofstream QOUT(f.c_str(),ios::out);
  QOUT.precision(4);
  
  QOUT << setw(par::pp_maxsnp) << "SNP" << " "
       << setw(4) << "CHR" << " " 
       << setw(4) << "N" << " "
       << setw(4) << "C0" << " "
       << setw(4) << "C1+" << " "
       << setw(10) << "FREQ" << " "
       << setw(10) << "BETA" << " "
       << setw(10) << "SE" << " "
       << setw(10) << "WALD" << " "
       << setw(10) << "P" << "\n";
  

  // Run original QTDT

  perm.setTests(nl_all);
  perm.setPermClusters(*this);
  
  vector_t orig = calcQTDT(true, QOUT, false, perm, pbetween, pwithin);

  QOUT.close();



  ////////////////
  // Permutation?
  
  // Not yet implemented...
  
  if (!par::permute) return;
  
  //////////////////////
  // Begin permutations
  
  bool finished = false;
  while(!finished)
    {

      ///////////////////////////////////
      // Set up permutation list for TDT
      
      permute(pbetween);
      
      // Within family permutation
      
      for (int i=0; i<family.size(); i++)
 	{
 	  if (CRandom::rand() < 0.5) pwithin[i] = true;
	  else pwithin[i] = false;
 	}
      
      
      // Recalculate QFAM tests
      
      vector<double> pr = calcQTDT(false,               // no output
				   QOUT,                // ignored 
				   true,                // permutation mode 
				   perm,
				   pbetween, pwithin);  // permutation vector (within-family)
      
      
      ////////////////////////////////
      // Standard permutation counting
      
      finished = perm.update(pr,orig);
      
    } // next permutation
  
  if (!par::silent)
    cout << "\n\n";
  
  
  ////////////////////
  // Display results
  
  ofstream TDT;   
  if (par::adaptive_perm) f = par::output_file_name + ".qfam.perm";
  else f = par::output_file_name + ".qfam.mperm";
  
  TDT.open(f.c_str(),ios::out);
  printLOG("Writing QFAM permutation results to [ " + f + " ] \n"); 
  TDT.precision(4);
  
  TDT << setw(4) << "CHR" << " "
      << setw(par::pp_maxsnp) << "SNP" << " ";
  
  if (par::perm_TDT_basic) TDT << setw(12) << "CHISQ" << " ";
  
  TDT << setw(12) << "EMP1" << " ";
  if (par::adaptive_perm)
    TDT << setw(12) << "NP" << " " << "\n";  
  else
    TDT << setw(12) << "EMP2" << " " << "\n";  
  
  
  for (int l=0; l<nl_all; l++)
    {	
      
      TDT << setw(4) << locus[l]->chr << " "
	  << setw(par::pp_maxsnp) << locus[l]->name << " "; 
      if (orig[l] < -0.5)
	TDT << setw(12) << "NA"  << " " << setw(12) << "NA"  << " " << setw(12) << "NA";
      else
	{
	  TDT << setw(12) << orig[l] << " "
	      << setw(12) << perm.pvalue(l) << " ";
	  
	  if (par::adaptive_perm)
	    TDT << setw(12) << perm.reps_done(l);
	  else
	    TDT << setw(12) << perm.max_pvalue(l);
	}
      TDT << "\n";
    }
  
  TDT.close();
    
   
}

vector_t Plink::calcQTDT(bool print_results, 
			 ofstream & QOUT,
			 bool permute, 
			 Perm & perm,
			 vector<int> & pbetween, 
			 vector<bool> & pwithin)
{
  
    
  //////////////////////
  // Dependent variable
  // and 
  // familial clustering

  vector_t Y; 
  vector<int> C;
  vector<Individual*>::iterator person = sample.begin(); 
  map<string,int> famcnt;
  int fc=0;
  while ( person != sample.end() )
  {
    if ( ! (*person)->missing )
     {
       Y.push_back( (*person)->phenotype );
       map<string,int>::iterator fi = famcnt.find( (*person)->fid );
       if ( fi == famcnt.end() )
	 {
	   famcnt.insert( make_pair( (*person)->fid , fc ));
	   C.push_back(fc);            
	   fc++;
	 }
       else
	 {
	   C.push_back( fi->second );
	 }   
     }
    
    person++;  
  } 
  int ns = Y.size();
  
  if (print_results)
    {
      printLOG(int2str(family.size())+" nuclear families in analysis\n");
      printLOG(int2str(famcnt.size())+" distinct family IDs found (clusters)\n");
    }

  if ( family.size()<2 )
    error("Halting: not enough nuclear families for this analysis\n");
  
 /////////////////////////
 // Iterate over each SNP

 vector_t results(nl_all);
 vector<bool> include;
 
 for (int l=0; l<nl_all; l++)
   {     

     if (par::verbose)
       cout << "\n ******************************************\n"
	    << "  LOCUS " << locus[l]->name << "\n\n";

     // In adaptive mode, possibly skip this test
     if (par::adaptive_perm && (!perm.test[l])) 
       continue; 
     
     // Form linear model
     
     LinearModel lm;
     
     vector<Individual*>::iterator gperson = sample.begin();
     
     vector_t X;
     vector<Individual*> tind;
     
     while ( gperson != sample.end() )
       {
	 
	 Individual * pperson = (*gperson)->pperson;
	 
	 if ( ! pperson->missing  )
	   {
	     double xi;
	     bool inc = true;

	     bool s1 = (*gperson)->one[l];
	     bool s2 = (*gperson)->two[l];	     
	     
	     if ( ! s1 ) 
	       {
		 if ( ! s2 ) 
		   xi = 1;
		 else 
		   xi = 0;
	       } 
	     else
	       {
		 if ( ! s2 ) 
		   inc = false;
		 else 
		   xi = -1;
	       }
	     
	     if (inc)
	       X.push_back(xi);
	     else
	       X.push_back(-9);
	     
	     include.push_back(inc);
	     tind.push_back(*gperson);
      
	   }
	 
	 // Next individual
	 gperson++;  	 
       }
     


     ///////////////////////////////////////
     // Score between and within components

     vector_t B;
     vector_t W;
     
     vector_t X2;
     vector_t Y2;
     vector<int> C2;
     
      vector<Family*>::iterator f = family.begin();
	 
     // Construct family B score for this SNP

     while ( f != family.end() )
       {
	     
	 if (par::verbose)
	   {
	     
	     if ( (*f)->singleton )
	       cout << "SINGLETON\t" 
		    << (*f)->kid[0]->fid << " " 
		    << (*f)->kid[0]->iid << "\n"; 
	     else if ( (*f)->sibship )
	       {
		 cout << "SIBSHIP " << (*f)->kid[0]->fid << " : ";
		 for ( int k=0; k<(*f)->kid.size(); k++)
		   cout << (*f)->kid[k]->iid << " ";
		 cout << "\n";
	       }
	     else if ( (*f)->parents )
	       {
		 cout << "W/ PARENTS " << (*f)->pat->fid << " ( ";
		 cout << (*f)->pat->iid << " x " << (*f)->mat->iid << " ) ";
		 for ( int k=0; k<(*f)->kid.size(); k++)
		   cout << (*f)->kid[k]->iid << " ";
		 cout << "\n";	     
	       }
	     else
	       cout << "WEIRD! " 
		    << (*f)->pat->fid << " "
		    << (*f)->pat->iid << "\n";
	     
	   }
	 	 	 
	 
	 double B = 0;
	 bool Bset = false;

	 (*f)->include = true;
	 
	 // Two genotyped parents?
	 if ( (*f)->parents )
	   {
	     
	     bool p1 = (*f)->pat->one[l];
	     bool p2 = (*f)->pat->two[l];
	     
	     bool m1 = (*f)->mat->one[l];
	     bool m2 = (*f)->mat->two[l];
	     
	     bool parents = true;
	     
	     if ( p1 && !p2 ) parents = false;
	     else if ( m1 && !m2 ) parents = false;
	     
	     // Use parental genotypes, if present for this SNP
	     if ( parents )
	       {
		 
		 if ( ! p1 )
		   {
		     if ( ! p2 ) 
		       B++;
		   }
		 else 
		   B--;
		 
		 
		 if ( ! m1 )
		   {
		     if ( ! m2 ) 
		       B++;
		   }
		 else 
		   B--;
		 
		 B /= 2;
		 Bset = true;
	       }
	   }
	 
	 if ( !Bset ) 
	   {
	     // Use sibling genotypes?

	     // (nb. this will default to own genotype, as singletons
	     //      are coded as offspring)
	     
	     int nsib = (*f)->kid.size();
	     int nsib2 = nsib;
	     for (int k=0; k<nsib; k++)
	       {
		 bool k1 = (*f)->kid[k]->one[l];
		 bool k2 = (*f)->kid[k]->two[l];
		 
		 if ( ! k1 ) 
		   {
		     if ( ! k2 )
		       B++;
		   }
		 else 
		   {
		     if ( ! k2 )
		       nsib2--;
		     else 
		       B--;
		   }		 
	       }
	     
	     if (nsib2==0)
	       {
		 // no non-missing offspring in family
		 // so does not matter what we set here
		 (*f)->include = false;		 
		 B = -9;
	       }
	     else
	       B /= (double)nsib2;
	   }
       	 	 
	 // Can remove this check
	 if ( ! realnum(B) ) error("problemmo!\n");

	 // Store between family score
	 (*f)->B = B;

	 // Next family
	 f++;
	 
       }
     

     vector<Individual*> tind2;
     gperson = sample.begin();
     int i=0;
     
     while ( gperson != sample.end() )
       {
	 
	 Individual * pperson = (*gperson)->pperson;
	 
	 // Valid phenotype...
	 if ( ( ! pperson->missing ) )
	   {
	     // ... and genotype?
	     if ( include[i] )
	       {
		 
		 Family * f = (*gperson)->family;
		 
		 // Are we modelling parental phenotypes?
		 if ( par::QFAM_total ||       // total association test
		      ( par::QFAM_within2 
 			&& f
			&& f->parents ) || // parenQTDT
		      ! (*gperson)->founder )  // not a founder
		   {
		     
		     // Between-family component
		     double b=0;
		     if ( f ) 
		       b = f->B;
		     
		     B.push_back( b );
		     
		     bool s1 = (*gperson)->one[l];
		     bool s2 = (*gperson)->two[l];	     
		     
		     if ( ! s1 ) 
		       {
			 if ( ! s2 ) 
			   W.push_back( 1 - b );
			 else 
			   W.push_back( 0 - b );
		       } 
		     else
		       {
			 if ( ! s2 ) 
			   W.push_back( -9 );
			 else
			   W.push_back( - 1 - b );
		       }
		     
		     // Add phenotype and genotype for extracted sample
		     Y2.push_back(Y[i]);
		     X2.push_back(X[i]);
		     C2.push_back(C[i]);
		     tind2.push_back(tind[i]);
		     
		     
		   }
	       }	     
	     
	     i++; // next non-missing person
	   }
	 
	 // Next person
	 gperson++;
	 
       }
     
       
     /////////////////////////////////////////
     // Permutation: flip betweens, withins
     // then reconstruct individual genotypes
     
     if (permute)
       {	 
	 
	 // Edit pbetween for this SNP, so that we keep missing 
	 // B components constant

	 vector<int> pb = pbetween;
	 
	 for (int f=0; f<pb.size(); f++)
	   {
	     
	     if ( 
		 // Permuted family is all missing
		 ( ! family[pb[f]]->include ) 
		 &&
		 // Recipient family is not...
		 family[f]->include )
	       {
		 // ... then swap 
		 
		 //   F  P(F)  
		 //   0  2
		 //   1  0
		 //   2  3*
		 //   3* 4
		 //   4  1
		 //   ...

		 // e.g. 3* is missing, so swap 3* and 4 in P(F), so 2
		 // and 4 end up together instead, 3* is invarint
		 
		 int missing_family = pb[f];
		 int swap_in_family = pb[pb[f]];
		 pb[missing_family] = missing_family;
		 pb[f] = swap_in_family;

// 		 cout << "FAM " << f << " (NOT MISS) has " << missing_family << " (MISS)\n";
// 		 cout << "FAM " << missing_family << " (MISS) has " << swap_in_family << " (?)\n";
// 		 cout << "SwAP MADE ..\n";
// 		 cout << "FAM " << f << "  has " << pb[f] << "\n";
// 		 cout << "FAM " << missing_family << " has " << pb[missing_family] << "\n\n";

		 // And re-check this new pairing
		 f--;
		 
	       }
	   }
	 
	 
 	 for (int i=0; i<X2.size(); i++)
 	   {
	     B[i] = family[pb[C2[i]]]->B;	 
 	     if ( pwithin[C2[i]] ) W[i] *= -1; 
 	     X2[i] = B[i] + W[i];
 	   }
       }


     if (par::verbose)
       {
	 cout << "Before pruneing Y,X,B,W sizes = "
	      << Y2.size() << " "
	      << X2.size() << " "
	      << B.size() << " "
	      << W.size() << " "
	      << tind2.size() << "\n";
	 
	 for (int i=0; i<Y2.size(); i++)
	   {
	     cout << tind2[i]->fid << " "
		  << tind2[i]->iid << "\t"
		  << C2[i] << "\t"
		  << Y2[i] << "\t"
		  << X2[i] << "\t"
		  << B[i] << "\t"
		  << W[i] << "\t";
	     
	     if ( ! family[C2[i]]->include ) cout << " EXCLUDED" ;
	     cout << "\n";
	   }
	 cout << "\n";
       }


     /////////////////////////
     // Prune out missing data
     
     Y.clear();
     X.clear();
     vector<int> C3;
     tind.clear();

     vector_t B2;
     vector_t W2;
     
     for (int i=0; i<Y2.size(); i++)
       {
	 // This individual has familial B component
	 if ( family[C2[i]]->include )
	   {
	     // and their own genotype data (-/+9 is W missing value)
	     if ( W[i] > -5 && W[i] < 5 )  
	       {
		 Y.push_back(Y2[i]);
		 X.push_back(X2[i]);
		 C3.push_back(C2[i]);
		 B2.push_back(B[i]);
		 W2.push_back(W[i]);
		 tind.push_back(tind2[i]);
	       }
	   }
       }
	 
	 
     /////////////////////////
     // Optional display


     if (par::verbose)
       {
	 cout << "After pruning Y,X,B,W sizes = "
	      << Y.size() << " "
	      << X.size() << " "
	      << C3.size() << " "
	      << B2.size() << " "
	      << W2.size() << " "
	      << tind.size() << "\n";
	 
	 for (int i=0; i<Y2.size(); i++)
	   cout << tind[i]->fid << " "
		<< tind[i]->iid << "\t"
		<< C3[i] << "\t"
		<< Y[i] << "\t"
		<< X[i] << "\t"
		<< B2[i] << "\t"
		<< W2[i] << "\n";
	 cout << "\n";
       }
     

     ////////////////////////
     // Perform QFAM analysis
     
     if (par::QFAM_total) 
       {
	 lm.setDependent(Y);
	 lm.noCluster();  
	 //lm.setCluster(C3);
	 lm.addIndependent(X);
       }
     else if (par::QFAM_within1 || par::QFAM_within2) 
       {
	 // Within-family QTDT, or parenQTDT
	 lm.setDependent(Y);
	 lm.noCluster();
	 //lm.setCluster(C3);
	 lm.addIndependent(W2);
	 lm.addIndependent(B2);
       }
 
     
     //////////////
     // Covariates?
     
     if (0) // par::covariates...
       {
	 // 
       } 

     lm.fitLM();

     vector_t b = lm.getCoefs();
     vector_t var = lm.getVar();

     // Test statistic
     results[l] = ( b[1]*b[1] ) / var[1] ;

     // Result?
     if (print_results)
       {
	 // Paremeter of interest is always b[1]
	 // (intercept b[0], between-family effects, covariates, b[2+]

	 double wald = ( b[1]*b[1] ) / var[1];
	 
	 QOUT << setw(par::pp_maxsnp) << locus[l]->name << " " 
	      << setw(4) << locus[l]->chr << " "
	      << setw(4) << X.size() << " "
	      << setw(4) << famcnt.size() << " "
	      << setw(10) << locus[l]->freq << " "
	      << setw(10) << b[1] << " "
	      << setw(10) << sqrt(var[1]) << " "
	      << setw(10) << wald << " "
	      << setw(10) << chiprobP(wald,1);
	 QOUT << "\n";
       }

     // Next SNP
   }

 return results;

}


